#include "DBSLib.h"
#include "DBSLibMem.h"
#include <Rect.h>				// for RectangleType
#include <Window.h>			// for WinDrawRectangle
#include "Debug.h"

// Define our library's version
#define dbsLibVersion	sysMakeROMVersion(1,0,0,sysROMStageRelease,0)



/***********************************************************************
 *
 *   Internal Functions
 *
 ***********************************************************************/
static Err ParseDBAttributes(Char *pszXML, Char *pszName, UInt32 *npDBType, UInt32 *npCreator);
static Err ParseFields(Char *chSchemaStart, Int16 *nNumFields, MemHandle *phFieldElemArray);
static Err ParseRecData(Char *pszStartDelim,
                        MemHandle hFieldElemArray, Int16 nNumFields, UInt32 *pnRecLen);
static Err CreateSchema(MemHandle hFieldElemArray, Int16 nNumElements, Char *pszSchema);
static void FreeFieldElemArray(MemHandle hFieldElemArray, Int16 nNumElements);
static void FreeFieldElemArrayData(MemHandle hFieldElemArray, Int16 nNumElements);
static Int16 CountItems(Char *pszStartDelim, Char* pszBeginTag, Char *pszEndTag);
static Err ExtractData(Char *pszString, Char *pszStartDelim, Char* pszEndDelim,
                       Int16 nMaxLen, Char *pszResult);
#if defined(DEBUG)
static void PrintFieldElementArray(MemHandle hFieldElemArray, Int16 nNumElements);
#endif


/************************************************************
 *
 *  FUNCTION: DBSLibOpen
 *
 *  DESCRIPTION:   Opens the DBSLib library, creates and initializes the globals.
 *                 This function must be called before any other Sample Library
 *                 functions.
 *
 *
 *  PARAMETERS:	refNum
 *              DBSLib library reference number returned by SysLibLoad() or
 *              SysLibFind().
 *
 *              clientContextPtr
 *              Pointer to variable for returning client context.
 *
 *************************************************************/
Err DBSLibOpen(UInt16 refNum, UInt32 * clientContextPtr)
{
   DBSLibGlobalMemPtr globalMemPtr;
   Err retVal = errNone;
   Int16 origRefCount = 0;

   TRACE_TL0("DBSLibOpen: IN");

   // Verify arguments
   ErrFatalDisplayIf(clientContextPtr == NULL, "DBSLib: NULL client context pointer");
		
   // Initialize return variable
   *clientContextPtr = 0;

   // Lock our global memory
   globalMemPtr = LockGlobalMem(refNum);
	
   // Check if our library is already open
   if (globalMemPtr == NULL)
   {
      // Create and initialize our global memory
      globalMemPtr = CreateGlobalMem(refNum);
      if (globalMemPtr == NULL )
         retVal = dbsLibErrMemory;
   }
	
   // If our library is open, create client memory and increment our reference
   // count
   if (globalMemPtr != NULL)
   {
      // save original open count for error handling
      origRefCount = globalMemPtr->libRefCount;
	
	  // create client memory
	  // upon success, *clientContextPtr will contain the client
	  // context.  the client application which opened our library
	  // needs to maintain this context until it closes our library	
      retVal = CreateClientMem(globalMemPtr, clientContextPtr);
      
      // increment the reference count upon success
      if (retVal == errNone)
         globalMemPtr->libRefCount++;
         
      // unlock our library globals
      UnlockGlobalMem(globalMemPtr);
		
      // If there was an error creating client memory and no other
      // clients have opened our library, free our globals
      if ( retVal != errNone && (origRefCount == 0) )
         FreeGlobalMem(refNum);
   }
	
   TRACE_TL0("DBSLibOpen: OUT");	
   return retVal;
}


/************************************************************
 *
 *  FUNCTION: DBSLibClose
 *
 *  DESCRIPTION:   Opens the DBSLib library, creates and initializes the globals.
 *                 This function must be called before any other Sample Library
 *                 functions.
 *
 *
 *  PARAMETERS:	refNum
 *              DBSLib library reference number returned by SysLibLoad() or
 *              SysLibFind().
 *
 *              clientContext
 *              Variable containing client context which was assigned in
 *              call to DBSLibOpen.  Used to access client memory.
 *
 *************************************************************/
Err DBSLibClose(UInt16 refNum, UInt32 clientContext)
{
   DBSLibGlobalMemPtr globalMemPtr;
   Int16 currLibCount;
   Int16 currCtxCount;
   Err retVal = errNone;

   TRACE_TL0("DBSLibClose: IN");
   
   //Lock our global memory
   globalMemPtr = LockGlobalMem(refNum);

   //Our library is already closed
   if(globalMemPtr == NULL)
      return retVal;
	
   // Free the client's memory
   // Freeing the client's memory automatically decrement's
   // the client context reference count
   FreeClientMem(globalMemPtr, clientContext);
	
   // Decrement our library's reference count
   globalMemPtr->libRefCount--;
	
   // Check for reference count underflow
   ErrFatalDisplayIf(globalMemPtr->libRefCount < 0, "DBSLib: Lib ref count underflow.");
	
	
   // Save the new library and context reference counts
   currLibCount = globalMemPtr->libRefCount;
   currCtxCount = globalMemPtr->clntRefCount;
	
   // Unlock our global memory
   UnlockGlobalMem(globalMemPtr);
	
   // If the library reference count has reached zero,
   // free our global memory
   if (currLibCount <= 0)
   {
      // Verify all client memory has been freed
      ErrFatalDisplayIf(currCtxCount != 0, "DBSLib: Not all client memory was freed.");
		
      // Free our global memory
      FreeGlobalMem(refNum);
   }
   else
   { 
      // return dbLibErrStillOpen to inform the caller
      // that other clients are still using our library
      retVal = dbsLibErrStillOpen;
   }
   
   TRACE_TL0("DBSLibClose: OUT");
   
   return retVal;
}

/************************************************************
 *
 *  FUNCTION: DBSLibSleep
 *
 *  DESCRIPTION: Handles OS sleep notification.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *************************************************************/
Err DBSLibSleep(UInt16 refNum)
{
	return errNone;
}

/************************************************************
 *
 *  FUNCTION: DBSLibWake
 *
 *  DESCRIPTION: Handles OS wake notification.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *************************************************************/
Err DBSLibWake(UInt16 refNum)
{
	return errNone;
}




/************************************************************
 *
 *  FUNCTION: DBSLibGetAPIVersion
 *
 *  DESCRIPTION: Get our library's API version.  Our library does not need to be opened
 *               to call this function.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *              pVer
 *              Filled with the version number of our library
 *
 *************************************************************/
Err DBSLibGetAPIVersion(UInt16 refNum, UInt32 *pVer)
{
	
	// Error-check our parameters
	ErrFatalDisplayIf(pVer == NULL, "DBSLib: Null pointer argument.");
	
	*pVer = dbsLibVersion;
	
	return errNone;
}

/************************************************************
 *
 *  FUNCTION: DBSLibParseSchema
 *
 *  DESCRIPTION: Populate a database with records from, and according
 *               to a schema from an XML string.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *              
 *              clientContext
 *              Variable containing client context which was assigned in
 *              call to DBSLibOpen.  Used to access client memory.
 *
 *              pszXML
 *              String containing database schema and data.
 *
 *              bDeleteDB
 *              If true, the database name specified in XML string
 *              is deleted before populating it with data.  If false,
 *              the database is not deleted; data is appended to it.
 *
 *************************************************************/
Err DBSLibParseSchema(UInt16 refNum, UInt32 clientContext,
                      Char* pszXML, Boolean bDeleteDB)
{

   Err retVal = errNone;
   //Start and end pointers for string parsing
   Char *pszStart, *pszEnd;
   Int16 nFieldIndex = 0, nRecIndex = 0;
   Int16 nNumFields = 0, nNumRecs = 0;      
   MemHandle hFieldElemArray = NULL;
   DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
   DBSLibClientMemPtr pClientMem;
      
   //Database attributes, as contained in the
   //XML string
   Char szDBName[DB_NAME_LEN];
   UInt32 nDBType = 0, nDBCreator = 0;
   
   
   //Variables for PalmOS database functions
   DmOpenRef dbOpenRef;         //A reference to the database we will open
   Boolean bIsResDB = false;    //we are not creating a resource database
   Boolean bIsDirty = true;     // Flag that our database has changed.
   UInt16 nDBMode = dmModeReadWrite; //Open database for read/write access
   UInt16 nCardNum = 0;              //Assume database is on card 0
   UInt16 nRecNum = 0;              //Current record being added
   UInt32 nRecLen = 0; //Length in bytes of data for each record
   UInt32 nRecOffset = 0;  //Offset into record to begin writing
   LocalID dbID; // local DB ID to determine if database exists
   MemHandle hRecord; //Handle to new record
   MemPtr pRecordLock; //Locked pointer to new record
   Boolean bCreateDB = false; //Create the database?
   
   TRACE_TL0("DBSLibParseSchema: IN");
   
   TRACE_TL1("pszXML = [%s]", pszXML);

   // Make sure the library has been opened
	if (IsDBSLibOpen(refNum) == false)
	   return dbsLibErrNotOpen;
      
   //Initialize the client's memory, including
   //state
   pClientMem = LockClientMem(clientContext);
   MemSet(pClientMem, sizeof(DBSLibClientMemType), 0);
   pClientMem->nState = dbsLibDBCreatePending;
   MemPtrUnlock(pClientMem);
   
   MemSet(szDBName, DB_NAME_LEN, 0);
 
    // Look for valid begin and end tags, and get the attributes
   retVal = ParseDBAttributes(pszXML, szDBName, &nDBType, &nDBCreator);      
   
   if(retVal != errNone)
   {
      TRACE_TL0("Error in call to ParseDBAttributes");
      return retVal;
   }     

   // Create the database
   // First see if it already exists
   dbID = DmFindDatabase(nCardNum, szDBName);
   
   if(dbID == 0)
   {
      bCreateDB = true;
   }
   else
   {
      //The database exists.  If we were told to delete it, do so.
      if(bDeleteDB == true)
      {
         retVal = DmDeleteDatabase(nCardNum, dbID);
         if(retVal != errNone)
         {
            TRACE_TL1("Could not delete %s database.", szDBName);
            return dbsLibErrDB;
         }
         else
         {
            TRACE_TL1("Deleted %s database.", szDBName);
         }
         bCreateDB = true;
      }
   }
       
   if(bCreateDB == true)
   {
      //The database doesn't exist, so create it
      retVal = DmCreateDatabase(nCardNum, (const Char*)szDBName, nDBCreator, nDBType, bIsResDB);                  
      if(retVal != errNone)
      {
         TRACE_TL1("Could not create %s database.", szDBName);
         return retVal;
      }
   }


   TRACE_TL0("Database:");
   TRACE_TL1("  Name:    %s", szDBName);
   TRACE_TL1("  Type:    %lx", nDBType);
   TRACE_TL1("  Creator: %lx", nDBCreator);

   //Parse the fields
   pszStart = StrStr(pszXML, "<Schema>");
   pszEnd = StrStr(pszXML, "</Schema>");
   
   if(pszStart == NULL || pszEnd == NULL)
      return dbsLibErrParse;
      
   pszStart += StrLen("<Schema>");
 
   //ParseFields parses the schema and returns an array of field elements
   //which describe the schema.
   retVal = ParseFields(pszStart, &nNumFields, &hFieldElemArray);
   
   if(retVal != errNone)
   {
      TRACE_TL0("Error in call to ParseFields");
      return retVal;
   }
      
   TRACE_TL1("nNumFields = %d", nNumFields);
   
   
   //Parse the records   
   pszStart = StrStr(pszXML, "<Data>");
   pszEnd = StrStr(pszXML, "</Data>");
   if(pszStart == NULL || pszEnd == NULL)
   {
      MemHandleFree(hFieldElemArray);
      return dbsLibErrParse;
   }
   pszStart += StrLen("<Data>");
   
   //Count the number of records between the <Data></Data> tags
   nNumRecs = CountItems(pszStart, "<Record>", "</Record>");
   
   TRACE_TL1("nNumRecs = %d", nNumRecs);

   if(nNumRecs == 0)
      return dbsLibErrParse;
         
   // Open the database
   dbOpenRef = DmOpenDatabaseByTypeCreator(nDBType, nDBCreator, nDBMode);

   if(dbOpenRef == 0)
   {
      FreeFieldElemArray(hFieldElemArray, nNumFields);
      return dbsLibErrDB;
   }  
   
   //For each record in the database, get its data from the XML string,
   //and add it to the database
   for(nRecIndex = 0; nRecIndex < nNumRecs; nRecIndex++)
   {
   
       TRACE_TL1("Adding record # %d", nRecIndex);
   
      //Find the <Record></Record> tags for this record's data
      pszStart = StrStr(pszStart, "<Record>");
      pszStart += StrLen("<Record>");
      pszEnd = StrStr(pszStart, "</Record>");
      

      //Retreive field data for given record
      retVal = ParseRecData(pszStart, hFieldElemArray, nNumFields, &nRecLen);
      
      if(retVal != errNone)
      {
         TRACE_TL0("Error in call to ParseRecData");
         FreeFieldElemArray(hFieldElemArray, nNumFields);
         DmCloseDatabase(dbOpenRef);
         return retVal;
      }
                 
     // Create a new database record for our data
     hRecord = DmNewRecord (dbOpenRef, &nRecNum, nRecLen);
     
     //Lock the record for editing
     pRecordLock = MemHandleLock(hRecord);
     
     //Lock the field element array, and add the elements
     //to the currently locked record
     pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);
     pCurrFieldElem = pFieldElemArray;
     nRecOffset = 0;
     for(nFieldIndex = 0; nFieldIndex < nNumFields; nFieldIndex++)     
     {
        void *pData;
        
        TRACE_TL2("Adding field %d to record %d", nFieldIndex, nRecIndex);

        //Lock the field element's data so we can write it
        //to the record
        pData = (void *)MemHandleLock(pCurrFieldElem->hData);
               
        //Write our data to the record
        retVal = DmWrite(pRecordLock, nRecOffset, pData, pCurrFieldElem->nDataLen);
        
        //Unlock the element's data
        MemPtrUnlock(pData);
                
        //Update the record offset and the field element
        //array pointer
        nRecOffset += pCurrFieldElem->nDataLen;
        pCurrFieldElem++;
     }
     MemPtrUnlock(pFieldElemArray);
          
     // Unlock the record handle
     MemPtrUnlock(pRecordLock);
     
     // Release the record
     DmReleaseRecord(dbOpenRef, nRecNum, bIsDirty);
     
     //Free this record's field element array data
     FreeFieldElemArrayData(hFieldElemArray, nNumFields);
               
   }
   
   //Close the database
   retVal = DmCloseDatabase(dbOpenRef);


   pClientMem = LockClientMem(clientContext);
   TRACE_TL0("Updating client memory.");
   StrCopy(pClientMem->szName, szDBName);
   //We can call CreateSchema with any instance of the field
   //element array.  In this case, we use the last instance
   CreateSchema(hFieldElemArray, nNumFields, pClientMem->szSchema);
   TRACE_TL1("Database schema = <%s>.", pClientMem->szSchema); 
   pClientMem->nNumFields = nNumFields;
   pClientMem->nNumRecords = nNumRecs;
   pClientMem->nState = dbsLibQuerySchemaPending;
   MemPtrUnlock(pClientMem);
  
   FreeFieldElemArray(hFieldElemArray, nNumFields);   

   TRACE_TL0("DBSLibParseSchema: OUT");

   return retVal;
}


/************************************************************
 *
 *  FUNCTION: DBSLibGetDBInfo
 *
 *  DESCRIPTION: Retrives information about the database just created
 *               according the the XML string passed into DBSLibParseSchema.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *              
 *              clientContext
 *              Variable containing client context which was assigned in
 *              call to DBSLibOpen.  Used to access client memory.
 *
 *              pszName
 *              Filled with the name of the database.
 *
 *              pnNumRecs
 *              Filled with the number of records added to the database in
 *              DBSLibParseSchema.
 *              ***NOTE:  This is NOT the same as the
 *                        total number of records in the database.
 *
 *              pNumFields
 *              Filled with the number of fields from the schema
 *              in the XML string passed into DBSLibParseSchema.
 *
 *              pszSchema
 *              Filled with the database schema, in a format suitable for
 *              parsing by an AppForge VB application.
 *
 *************************************************************/
Err DBSLibGetDBInfo(UInt16 refNum, UInt32 clientContext, Char* pszName, Int16 *pnNumRecs, Int16 *pnNumFields, Char* pszSchema)
{
   DBSLibClientMemPtr pClientMem;

   // Make sure the library has been opened
	if (IsDBSLibOpen(refNum) == false)
	   return dbsLibErrNotOpen;
   
   //Lock the client's memory
   pClientMem = LockClientMem(clientContext);
   
   //Check the state to make sure we have
   //successfully created a database
   if(pClientMem->nState != dbsLibQuerySchemaPending)
      return dbsLibErrState;
      
   //Fill the arguments with the requested data
   StrCopy(pszName, pClientMem->szName);
   *pnNumRecs = pClientMem->nNumRecords;
   *pnNumFields = pClientMem->nNumFields;   
   StrCopy(pszSchema, pClientMem->szSchema);
   
   //Unlock the client's memory
   MemPtrUnlock(pClientMem);
   
   return errNone;
}


/************************************************************
 *
 *  FUNCTION: ParseFields
 *
 *  DESCRIPTION: Parses the fields in an XML string, creating an array
 *               of field type elements which describe how each record
 *               in the database is formatted.
 *
 *  PARAMETERS: chSchemaStart
 *             
 *
 *              pszName
 *              Filled with the name of the database.
 *
 *              pnNumRecs
 *              Filled with the number of records added to the database in
 *              DBSLibParseSchema.  Note:  This is NOT the same as the
 *              total number of records in the database.
 *
 *              pNumFields
 *              Filled with the number of fields from the schema
 *              in the XML string passed into DBSLibParseSchema.
 *
 *              pszSchema
 *              Filled with the database schema, in a format suitable for
 *              parsing by an AppForge VB application.
 *
 *************************************************************/
Err ParseFields(Char* pszSchemaStart, Int16* npNumFields, MemHandle *phFieldElemArray)
{
   Err retVal = errNone;
   Boolean bContinue = true;
   Char *pszStart, *pszEnd;
   Int16 nIndex;
   MemHandle hFieldElemArray;
   DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
   Int16 nFieldElemTypeSize;
   UInt32 nFieldArraySize;
   UInt16 nResult;
    
   Char szTemp[32];

   TRACE_TL0("ParseFields: IN");
   
   TRACE_TL1("pszSchemaStart = [%s]", pszSchemaStart);

   *npNumFields = 0;
   *phFieldElemArray = NULL;
   
   *(npNumFields) = CountItems(pszSchemaStart, "<Field>", "</Field>");
   
   TRACE_TL1("Num fields = %d", (*npNumFields));

   //There has to be at least one field
   if(*(npNumFields) < 1)
      return dbsLibErrParse;

     
   //Now that we have the number of fields, we need to allocate memory
   //for the field array and parse the field types.
   
   nFieldElemTypeSize = sizeof(DBSLibFieldElemType);
   TRACE_TL1("Field element size = %d", nFieldElemTypeSize);
   nFieldArraySize = (*npNumFields)*nFieldElemTypeSize;
   TRACE_TL1("Field array size = %lu", nFieldArraySize);
   hFieldElemArray = MemHandleNew(nFieldArraySize);
   if(hFieldElemArray == 0)
   {
      TRACE_TL0("Could not allocate memory for field array");
      return dbsLibErrMemory;
   }
      
   //Lock the field array for modification
   pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);
   
   //Initialize field array to 0
   MemSet(pFieldElemArray, (*npNumFields)*sizeof(DBSLibFieldElemType), 0);
   
   
   //Loop for the number of fields, filling the array
   pszStart = pszSchemaStart;
   pCurrFieldElem = pFieldElemArray;
   for(nIndex = 0; nIndex < *(npNumFields); nIndex++)
   {
      //Find beginning field tag
      pszStart = StrStr(pszStart, "<Field>");
      if(pszStart == NULL)
      {
         TRACE_TL0("Did not find <Field> begin tag.");
         MemPtrFree(pFieldElemArray);
         return dbsLibErrParse;
      }
      
      pszStart += StrLen("<Field>");
      
      //Find end field tag
      pszEnd = StrStr(pszStart, "</Field>");
      if(pszEnd == NULL)
      {
         TRACE_TL0("Did not find <Field> end tag.");
         MemPtrFree(pFieldElemArray);
         return dbsLibErrParse;
      }

      //Extract the data between the <Num></Num> tag pair      
      MemSet(szTemp, 32, 0);
      nResult = ExtractData(pszStart, "<Num>", "</Num>", 31, szTemp);
      if(nResult != errNone)
      {
         MemPtrFree(pFieldElemArray);
         return nResult;
      }
      
      TRACE_TL1("Num: szTemp = %s", szTemp);
      pCurrFieldElem->nNum = StrAToI(szTemp);

      //Extract the data between the <Name></Name> tag pair
      MemSet(pCurrFieldElem->szName, DB_NAME_LEN, 0);
      nResult = ExtractData(pszStart, "<Name>", "</Name>", DB_NAME_LEN-1, pCurrFieldElem->szName);
      if(nResult != errNone)
      {
         MemPtrFree(pFieldElemArray);
         return nResult;
      }    
      

      //Extract the data between the <Type></Type> tag pair
      MemSet(szTemp, 32, 0);
      nResult = ExtractData(pszStart, "<Type>", "</Type>", 31, szTemp);
      if(nResult != errNone)
      {
         MemPtrFree(pFieldElemArray);
         return nResult;
      }
          
      pCurrFieldElem->nType = StrAToI(szTemp);
      
      TRACE_TL1("pCurrFieldElem->nType = %d", pCurrFieldElem->nType); 
    
       
      //Extract  the data between the <Length></Length> tag pair
      MemSet(szTemp, 32, 0);
      nResult = ExtractData(pszStart, "<Length>", "</Length>", 31, szTemp);
      if(nResult != errNone)
      {
         MemPtrFree(pFieldElemArray);
         return nResult;
      }
            
      pCurrFieldElem->nLength = StrAToI(szTemp);
      
      TRACE_TL1("pCurrFieleElem->nLength = %d", pCurrFieldElem->nLength);

      
      //Increment the array pointer to the
      //next element in the array 
      pCurrFieldElem++;
   }
   
   //Assign the argument passed in to the
   //allocated array
   (*phFieldElemArray) = hFieldElemArray;
   
   //Unlock the array's memory
   MemPtrUnlock(pFieldElemArray);
         
         
   TRACE_TL0("ParseFields: OUT");
   
   return retVal;
}

/************************************************************
 *
 *  FUNCTION: CountItems
 *
 *  DESCRIPTION: Count the number of begin and end tags found in a
 *               given string.  For example, find the number of times
 *               the <Record></Record> tag pair is found within a
 *               string.
 *
 *  PARAMETERS:	pszStartDelim
 *              Start of string to parse.
 *              
 *              pszBeginTag
 *              Beginning XML tag to look for
 *
 *              pszEndTag
 *              Ending XML tag to look for.
 *
 *************************************************************/
Int16 CountItems(Char *pszStartDelim, Char* pszBeginTag, Char *pszEndTag)
{
   Char *pszStart = pszStartDelim, *pszEnd;
   Int16 nCount = 0;
   Boolean bContinue = true;
   
   while(bContinue == true)
   {
      //Find the begin tag
      pszStart = StrStr(pszStart, pszBeginTag);
      
      //Find the end tag
      if(pszStart != NULL)
         pszEnd = StrStr(pszStart, pszEndTag);
      
      if(pszStart == NULL || pszEnd == NULL)
      {
         //Assume there are no more fields
         bContinue = false;
      }
      else
      {
         //Skip past the begin tag
         pszStart += StrLen(pszBeginTag);
         
         //Increment the count
         nCount++;
      }
      
   }   
   
   return nCount;
}

/************************************************************
 *
 *  FUNCTION: ParseDBAttributes
 *
 *  DESCRIPTION: Parse an XML string for "Database" tags, and extract
 *               the database name, type, and creator attributes.
 *
 *  PARAMETERS:	pszXML
 *              Start of XML string to parse.
 *              
 *              pszDBName
 *              Filled with the database name to create.
 *
 *              pnDBType
 *              Filled with the type of database.
 *
 *              pnDBCreator
 *              Filled with the creator ID for this database.
 *
 *************************************************************/
Err ParseDBAttributes(Char *pszXML, Char *pszDBName, UInt32 *pnDBType, UInt32 *pnDBCreator)
{
   Err retVal = errNone;
   Char *pszStart, *pszEnd;
   UInt16 nResult;
   Char szTemp[16];
      
   if(pszXML == NULL)
      return dbsLibErrArg;

   //Find the start and end tags
   pszStart = StrStr(pszXML, "<Database");
   pszEnd = StrStr(pszXML, "</Database>");
   
   //If these tags are missing, return an
   //error
   if(pszStart == NULL || pszEnd == NULL)
      return dbsLibErrParse;
   
   //Extract the Name attribute
   nResult = ExtractData(pszStart, "Name=", " ", DB_NAME_LEN, pszDBName);
   if(nResult != errNone)
      return nResult;
      
 
   //Extract the Type attribute
   MemSet(szTemp, sizeof(szTemp), 0);
   nResult = ExtractData(pszStart, "Type=", " ", sizeof(*pnDBType), szTemp);
   if(nResult != errNone)
      return nResult;
      
   //Assign the Type to the argument
   *pnDBType = (*(UInt32*)szTemp);
   


   //Extract the Creator attribute
   MemSet(szTemp, sizeof(szTemp), 0);
   nResult = ExtractData(pszStart, "Creator=", ">", sizeof(*pnDBCreator), szTemp);
   if(nResult != errNone)
      return nResult;
   
   //Assign the Creator to the argument
   *pnDBCreator = (*(UInt32*)szTemp);
   

   return retVal;
}


/************************************************************
 *
 *  FUNCTION: ParseRecData
 *
 *  DESCRIPTION: Given a field element array defining a set of fields
 *               for a record, parse an XML string for each field, and
 *               extract its data. For each element in the array, search
 *               for data between the tags <NAME></NAME>, where NAME is
 *               the name of that element's data.
 *
 *  PARAMETERS:	pszStartDelim
 *              Start of XML string to parse.
 *              
 *              hFieldElemArray
 *              Handle to field element array.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *              pnRecRecLen
 *              Filled with the length of the record's data.  The record's
 *              data consists of the sum of the length of each field's
 *              data in the record.
 *
 *************************************************************/
Err ParseRecData(Char *pszStartDelim, MemHandle hFieldElemArray, Int16 nNumFields, UInt32 *pnRecLen)
{
  DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
  Int16 nFieldIndex;
  Char *pszTemp;
  Char szBeginTag[64];
  Char szEndTag[64];
  Char szData[MAX_DATA_LEN];
  UInt16 nResult;
  
  
  TRACE_TL0("ParseRecData: IN");
  
  pFieldElemArray = pCurrFieldElem = NULL;
  *pnRecLen = 0;
  
  //For each field in the array, look for a corresponding record entry
  pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);

     
  pCurrFieldElem = pFieldElemArray;
  for(nFieldIndex = 0; nFieldIndex < nNumFields; nFieldIndex++)
  {
     StrCopy(szBeginTag, "<");
     StrCat(szBeginTag, pCurrFieldElem->szName);
     StrCat(szBeginTag, ">");
     pszTemp = szBeginTag;
     pszTemp++;
     StrCopy(szEndTag, "</");
     StrCat(szEndTag, pszTemp);
        
     //Extract the data between the two tags
     MemSet(szData, MAX_DATA_LEN, 0);
     nResult = ExtractData(pszStartDelim, szBeginTag, szEndTag, MAX_DATA_LEN-1, szData); 
     if(nResult != errNone)
        return nResult;
        
               
     switch(pCurrFieldElem->nType)
     {
        case F_TYPE_STR:   //Char
           pCurrFieldElem->nDataLen = StrLen(szData) + 1;
        break;
        
        case F_TYPE_INT:   //Int16
           pCurrFieldElem->nDataLen = sizeof(Int16);
           TRACE_TL1("Int length = %d", pCurrFieldElem->nDataLen);
        break;
        
        case F_TYPE_LONG:  //Int32
           pCurrFieldElem->nDataLen = sizeof(Int32);
           TRACE_TL1("Long length = %d", pCurrFieldElem->nDataLen);
        break;
        
        default:
           return dbsLibErrParse;
        break;
     }
     
     *pnRecLen += pCurrFieldElem->nDataLen;
      
     //Create the data
     pCurrFieldElem->hData = MemHandleNew(pCurrFieldElem->nDataLen);
     if(pCurrFieldElem->hData == NULL)
     {
        TRACE_TL0("Could not create memory for field element data");
        return dbsLibErrMemory;
     }
     
     switch(pCurrFieldElem->nType)
     {
        case F_TYPE_STR:   //Char
        {
           Char* pData = (Char *)MemHandleLock(pCurrFieldElem->hData);
           if(pData == NULL)
              return dbsLibErrMemory;
           
           MemSet(pData, pCurrFieldElem->nDataLen, 0);
           StrCopy(pData, szData);
           MemPtrUnlock(pData);
        }
        break;
        
        case F_TYPE_INT:   //Int16
        {
           Int16* pData = (Int16 *)MemHandleLock(pCurrFieldElem->hData);
           if(pData == NULL)
              return dbsLibErrMemory;
           
           MemSet(pData, pCurrFieldElem->nDataLen, 0);
           *pData = StrAToI(szData);
           MemPtrUnlock(pData);
            
        }
        break;
        
        case F_TYPE_LONG:  //Int32
        {
           Int32* pData = (Int32 *)MemHandleLock(pCurrFieldElem->hData);
           if(pData == NULL)
              return dbsLibErrMemory;
           
           MemSet(pData, pCurrFieldElem->nDataLen, 0);
           *pData = StrAToI(szData);
           MemPtrUnlock(pData);
            
        }
        break;
        
        default:
           return dbsLibErrParse;
        break;
     }
          
     pCurrFieldElem++;
  }

   
  //Unlock the field element array
  MemPtrUnlock(pFieldElemArray);

  #if defined(DEBUG)
  PrintFieldElementArray(hFieldElemArray, nNumFields);
  #endif

  TRACE_TL0("ParseRecData: OUT");


  return errNone;
}

/************************************************************
 *
 *  FUNCTION: PrintFieldElementArray
 *
 *  DESCRIPTION: Debug utility for printing the elements of a
 *               field element array.
 *
 *  PARAMETERS:	hFieldElemArray
 *              Handle to field element array.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *************************************************************/
#if defined(DEBUG)
void PrintFieldElementArray(MemHandle hFieldElemArray, Int16 nNumElements)
{
   DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
   Int16 nFieldIndex;
   
   TRACE_TL0("PrintFieldElementArray: IN");
   
   pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);
   
   if(pFieldElemArray == NULL)
      return;
      
   pCurrFieldElem = pFieldElemArray;
   for(nFieldIndex = 0; nFieldIndex < nNumElements; nFieldIndex++)
   {
      TRACE_TL1("Field element %d:", nFieldIndex);
      TRACE_TL1("     Num:   %ld", pCurrFieldElem->nNum);
      TRACE_TL1("    Name:   %s", pCurrFieldElem->szName);
      
      switch(pCurrFieldElem->nType)
      {
         case  F_TYPE_STR: //Char
         {
            Char *pData;
            TRACE_TL0("    Type:   String");
            TRACE_TL1(" DataLen:   %lu", pCurrFieldElem->nDataLen);
            pData = (Char *)MemHandleLock(pCurrFieldElem->hData);
            TRACE_TL1("    Data:   %s", pData);
            MemPtrUnlock(pData);
         }
         break;
         
         case F_TYPE_INT:  //Int16
         {
            Int16 *pData;
            TRACE_TL0("    Type:   Int");
            TRACE_TL1(" DataLen:   %lu", pCurrFieldElem->nDataLen);
            pData = (Int16 *)MemHandleLock(pCurrFieldElem->hData);
            TRACE_TL1("    Data:   %d", *pData);
            MemPtrUnlock(pData);
         }
         break;
         
         case F_TYPE_LONG:  //Int32
         {
            Int32 *pData;
            TRACE_TL0("    Type:   Long");
            TRACE_TL1(" DataLen:   %lu", pCurrFieldElem->nDataLen);
            pData = (Int32 *)MemHandleLock(pCurrFieldElem->hData);
            TRACE_TL1("    Data:   %ld", *pData);
            MemPtrUnlock(pData);
         }
         break;
         
         default:
            TRACE_TL0("    Type:   Unknown");
         break;
      }
      
      TRACE_TL1("  Length:   %ld", pCurrFieldElem->nLength);
      
      pCurrFieldElem++;
   }
   
   MemPtrUnlock(pFieldElemArray);
   
   TRACE_TL0("PrintFieldElementArray: OUT");
    
}
#endif

/************************************************************
 *
 *  FUNCTION: FreeFieldElemArray
 *
 *  DESCRIPTION: Frees a field element array.
 *
 *  PARAMETERS:	hFieldElemArray
 *              Handle to field element array, which is freed.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *************************************************************/
void FreeFieldElemArray(MemHandle hFieldElemArray, Int16 nNumElements)
{
   //Free each element's data first
   FreeFieldElemArrayData(hFieldElemArray, nNumElements);
   
   //Now free the array
   MemHandleFree(hFieldElemArray);
   
}

/************************************************************
 *
 *  FUNCTION: FreeFieldElemArrayData
 *
 *  DESCRIPTION: Frees the data contained in the field element array.
 *
 *  PARAMETERS:	hFieldElemArray
 *              Handle to field element array, for which each element's
 *              data is freed.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *************************************************************/
void FreeFieldElemArrayData(MemHandle hFieldElemArray, Int16 nNumElements)
{
   DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
   Int16 nIndex;

   pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);
   pCurrFieldElem = pFieldElemArray;     
   for (nIndex = 0; nIndex < nNumElements; nIndex++)
   {
      // Free the field's data, if it has been
      // allocated
      if(pCurrFieldElem->hData != NULL)
         MemHandleFree(pCurrFieldElem->hData);
         
      pCurrFieldElem->hData = NULL;
      
      pCurrFieldElem++;
   }
   
   MemPtrUnlock(pFieldElemArray);

}

/************************************************************
 *
 *  FUNCTION: CreateSchema
 *
 *  DESCRIPTION:   Creates a schema of the form "field_num field_type field_len ..."
 *                 using the array of field elements which is passed in.
 *
 *  PARAMETERS:	hFieldElemArray
 *              Handle to field element array, from which the schema is created.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *              pszSchema
 *              Caller allocated string of length DB_SCHEMA_LEN.
 *              Resultant schema is stored in this argument.
 *
 *************************************************************/
Err CreateSchema(MemHandle hFieldElemArray, Int16 nNumElements, Char *pszSchema)
{
   DBSLibFieldElemTypePtr pFieldElemArray, pCurrFieldElem;
   Int16 nFieldIndex;
   Char szTemp[16];
   Int16 nSchemaLen;
   
   pFieldElemArray = (DBSLibFieldElemTypePtr)MemHandleLock(hFieldElemArray);
   
   pCurrFieldElem = pFieldElemArray;
   nSchemaLen = 0;   
   for(nFieldIndex = 0; nFieldIndex < nNumElements; nFieldIndex++)
   {
      
      //Schema string is of the form:
      //field_num field_type field_len ...
      StrPrintF(szTemp, "%ld %d %ld ",
                        pCurrFieldElem->nNum,
                        pCurrFieldElem->nType,
                        pCurrFieldElem->nLength);
      nSchemaLen += StrLen(szTemp);
      if(nSchemaLen > DB_SCHEMA_LEN)
      {
         MemPtrUnlock(pFieldElemArray);
         return dbsLibErrParse;
      } 
      StrCat(pszSchema, szTemp);
      pCurrFieldElem++;
   }
   
   return errNone;
}

/************************************************************
 *
 *  FUNCTION: ExtractData
 *
 *  DESCRIPTION: Extracts data (characters) from a string which
 *               occurs between a start and end delimiter.
 *
 *  PARAMETERS:	hFieldElemArray
 *              Handle to field element array, from which the schema is created.
 *
 *              nNumElements
 *              Number of elements in the field element array.
 *
 *              pszSchema
 *              Caller allocated string of length DB_SCHEMA_LEN.
 *              Resultant schema is stored in this argument.
 *
 *************************************************************/
Err ExtractData(Char *pszString, Char *pszStartDelim, Char* pszEndDelim, Int16 nMaxLen, Char *pszResult)
{

   Char *pszStart, *pszEnd;
   Int16 nIndex;
  
   TRACE_TL0("ExtractData: IN");
   TRACE_TL1("String = %s", pszString);
   TRACE_TL2("start = %s, end = %s", pszStartDelim, pszEndDelim);
     
   //Find the beginning delimiter
   pszStart = StrStr(pszString, pszStartDelim);
  
   if(pszStart == NULL)
   {
      TRACE_TL0("pszStart = NULL");
      return dbsLibErrParse;
   }
  
   //Increment past the start tag
   pszStart += StrLen(pszStartDelim);
  
   pszEnd = StrStr(pszStart, pszEndDelim);
  
   if(pszEnd == NULL)
   {
      TRACE_TL0("pszEnd = NULL");
      return dbsLibErrParse;
   }
     
   //Extract the data within the tags
   nIndex = 0;
   //Loop through the string, extracting each character until we
   //reach the end delimiter or exceed the maximum length
   while(StrNCompare(pszStart, pszEndDelim, StrLen(pszEndDelim)) != 0)
   {
      pszResult[nIndex] = *pszStart;
      nIndex++;
      pszStart++;
     
      if(nIndex > nMaxLen)
      {
         TRACE_TL1("nIndex of %d is > nMaxLen", nIndex);
         return dbsLibErrParse;
      }
   }
  
   TRACE_TL1("pszResult = %s", pszResult);
      
   TRACE_TL0("ExtractData: OUT");
   
   return errNone;
}