#include <PalmOS.h>

#include "DBSLib.h"
#include "DBSLibMem.h"
#include "Debug.h"


/************************************************************
 *
 *  FUNCTION: CreateGlobalMem
 *
 *  DESCRIPTION: Create our library's global memory
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *************************************************************/
DBSLibGlobalMemPtr CreateGlobalMem(UInt16 refNum)
{
    // Pointers and handles to our library's global memory
	DBSLibGlobalMemPtr globalMemPtr = NULL;							
	MemHandle globalMemHandle;
	// Pointer to our library's table entry as maintained by the OS
	SysLibTblEntryPtr libEntryPtr;


	// Get our library's sytem library table entry
	libEntryPtr = SysLibTblEntry(refNum);			
	ErrFatalDisplayIf(libEntryPtr == NULL, "DBSLib: Invalid refNum.");
	
	// Make sure our global memory doesn't already exist
	ErrFatalDisplayIf(libEntryPtr->globalsP != NULL, "DBSLib: Globals already exist.");
	
	
	// Allocate and initialize our global memory.
    // NOTE:  Do not use MemPtrNew, as this locks down our global
    //        memory, which could increase heap fragmentation.
	globalMemHandle = MemHandleNew(sizeof(DBSLibGlobalMemType));
	if (globalMemHandle == NULL)
		return NULL;     // could not allocate memory
	
	// Save the handle of our global memory in the system library table entry,
	// so we can access it later.
	// NOTE:  Again, although globalsP is declared as a MemPtr,
	//        you should not use MemPtrNew and assign the return
	//        value to globalsP, since MemPtrNew locks down our
	//        global memory, which could increase heap fragmentation.
	libEntryPtr->globalsP = (void*)globalMemHandle;


	// Go ahead and lock our globals; the calling routine
	// is responsible for unlocking them
	globalMemPtr = LockGlobalMem(refNum);
	ErrFatalDisplayIf(globalMemPtr == NULL, "DBSLib: Failed to lock globals.");
	
	// Assign ownership of our global memory to the OS so Palm's memory manager
	// will not automatically reclaim our it as client applications close our
	// library.
	// We will free our own memory, when our library's reference count goes to 0.
	MemPtrSetOwner(globalMemPtr, 0);
	
	// Initialize our global memory
	MemSet(globalMemPtr, sizeof(DBSLibGlobalMemType), 0);
	globalMemPtr->refNum = refNum;
	//initial open count is 0
	globalMemPtr->libRefCount = 0;
	//initial client count is 0
	globalMemPtr->clntRefCount = 0;
	
	// Return a pointer to our library's globals, which are locked
	// The calling routine is responsible for unlocking them
	return globalMemPtr;

}

/************************************************************
 *
 *  FUNCTION: FreeGlobalMem
 *
 *  DESCRIPTION: Free our library's global memory.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number. 
 *
 *************************************************************/
void FreeGlobalMem(UInt16 refNum)
{
	// Handle to our library's global memory
	MemHandle globalMemHandle;
	// Pointer to our library's table entry as maintained by the OS
	SysLibTblEntryPtr libEntryPtr;

    //Calling SysLibTblEntry with our refNum gets our system library table entry.
    //From this entry, we can get the handle to our global memory.
	libEntryPtr = SysLibTblEntry(refNum);
	ErrFatalDisplayIf(libEntryPtr == NULL, "DBSLib: Invalid refNum.");
	globalMemHandle = (MemHandle)(libEntryPtr->globalsP);			
	
	// Free our library's global memory
	if(globalMemHandle != NULL)
    {
       TRACE_TL0("Freeing global memory");
       //Clear our global pointer
       libEntryPtr->globalsP = NULL;
       //Free our global memory
	   MemHandleFree(globalMemHandle);
	}

}

/************************************************************
 *
 *  FUNCTION: LockGlobalMem
 *
 *  DESCRIPTION: Lock our library's global memory.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *************************************************************/
DBSLibGlobalMemPtr LockGlobalMem(UInt16 refNum)
{
    // Pointers and handles to our library's global memory
	DBSLibGlobalMemPtr globalMemPtr = NULL;							
	MemHandle globalMemHandle;
	// Pointer to our library's table entry as maintained by the OS
	SysLibTblEntryPtr libEntryPtr;

    //Calling SysLibTblEntry with our refNum gets our system library table entry.
    //From this entry, we can get the handle to our global memory.
	libEntryPtr = SysLibTblEntry(refNum);
	if (libEntryPtr != NULL)
		globalMemHandle = (MemHandle)(libEntryPtr->globalsP);
	
	//Lock our global memory	
	if (globalMemHandle != NULL)
		globalMemPtr = (DBSLibGlobalMemPtr)MemHandleLock(globalMemHandle);

	return globalMemPtr;

}


/************************************************************
 *
 *  FUNCTION: IsDBSLibOpen
 *
 *  DESCRIPTION: Determine if our library is open.
 *               This is done by attempting to lock
 *               our library's global memory.
 *
 *  PARAMETERS: refNum
 *              DBSLib reference number.
 *
 *************************************************************/
Boolean IsDBSLibOpen(UInt16 refNum)
{
   DBSLibGlobalMemPtr	globalMemPtr;
   Boolean	isOpen = false;
	
	globalMemPtr = LockGlobalMem(refNum);
	
	if (globalMemPtr != NULL)
	{
	   isOpen = true;
	   UnlockGlobalMem(globalMemPtr);
	}
	
	return isOpen;
}


/************************************************************
 *
 *  FUNCTION: CreateClientMem
 *
 *  DESCRIPTION: Create client memory.
 *
 *  PARAMETERS: globalMemPtr
 *              A locked pointer to our library's global memory.
 *
 *              clientContextPtr
 *              Filled with the handle to the client's allocated
 *              memory.  This handle becomes the client context,
 *              which needs to be passed into any custom library
 *              function which accesses client memory.
 *
 *************************************************************/
Err CreateClientMem(DBSLibGlobalMemPtr globalMemPtr, UInt32 * clientContextPtr)
{
   Err retVal = errNone;
   MemHandle clientMemHandle;
   DBSLibClientMemPtr clientMemPtr;

	
   // Verify the parameters
   ErrFatalDisplayIf(globalMemPtr == NULL, "DBSLib: NULL global memory pointer.");
   ErrFatalDisplayIf(clientContextPtr == NULL, "DBSLib: NULL client context pointer.");
	
	
   // Initialize client context
   *clientContextPtr = NULL;
	

   // Allocate a new client memory structure
   clientMemHandle = MemHandleNew(sizeof(DBSLibClientMemType));
   if(clientMemHandle == NULL)
      retVal = dbsLibErrMemory;
   else
   {
      // Save the client's memory handle in the client
      // context passed into this routine
      *clientContextPtr = (UInt32)clientMemHandle;
		
      // Lock and initialize the client memory
      clientMemPtr = (DBSLibClientMemPtr)MemHandleLock(clientMemHandle);
      MemSet(clientMemPtr, sizeof(DBSLibClientMemType), 0);
		
      //Unlock the client's memory
      UnlockClientMem(clientMemPtr);

      //Increase the client reference count in global memory
      globalMemPtr->clntRefCount++;
		
   }

   return retVal;
}

/************************************************************
 *
 *  FUNCTION: FreeClientMem
 *
 *  DESCRIPTION: Free client memory
 *
 *  PARAMETERS: globalMemPtr
 *              A locked pointer to our library's global memory.
 *
 *              clientContext
 *              Contains the handle to the client's allocated
 *              memory.
 *
 *************************************************************/
Err FreeClientMem(DBSLibGlobalMemPtr globalMemPtr, UInt32 clientContext)
{
   DBSLibClientMemPtr clientMemPtr;
	
   // Error-check our parameters
   ErrFatalDisplayIf(globalMemPtr == NULL, "DBSLib: NULL global memory pointer.");

   // Validate the client context by locking it
   clientMemPtr = LockClientMem(clientContext);
	
   if(clientMemPtr != NULL)
   {
      //MemPtrFeee can be called on a locked chunk
      MemPtrFree(clientMemPtr);
      //Decrement the client reference count
      globalMemPtr->clntRefCount--;

      ErrFatalDisplayIf(globalMemPtr->clntRefCount < 0, "DBSLib: client ref count underflow.");
   }
	
   return errNone;
}

/************************************************************
 *
 *  FUNCTION: LockClientMem
 *
 *  DESCRIPTION: Lock client memory.
 *
 *  PARAMETERS: clientContext
 *              Contains the handle to the client's allocated
 *              memory.
 *
 *************************************************************/
DBSLibClientMemPtr LockClientMem(UInt32 clientContext)
{
   DBSLibClientMemPtr clientMemPtr = NULL;
	
	
   // Verify the parameters
   ErrFatalDisplayIf(clientContext == NULL, "DBSLib: NULL client context.");
	
   // Lock the client's memory
   clientMemPtr = (DBSLibClientMemPtr)MemHandleLock((MemHandle)clientContext);
	
   ErrFatalDisplayIf(clientMemPtr == NULL, "DBSLib: Failed to lock client memory.");
	
   return clientMemPtr;
}